module Flash
  ( Flash(..)
  , compile
  )
 where

import Lava

----------------------------------------------------------------
-- Flash datatype

data Flash out
  = Skip
  | Emit out
  | Wait
  | IfThenElse (Signal Bool) (Flash out, Flash out)
  | While (Signal Bool) (Flash out)
  | Flash out :>> Flash out
  | Flash out :|| Flash out

-- for the translation

type FlashEmits out
  = [(out, Signal Bool)]

type FlashCircuit out
  = Signal Bool -> (FlashEmits out, Signal Bool)

----------------------------------------------------------------
-- compilation

compile :: Eq out => Flash out -> [out] -> [Signal Bool]
compile prog outputs = signals
  where
    (emits, _) = circuit prog start
    start      = delay high low
    
    signals =
      [ orl [ sig
            | (out',sig) <- emits
            , out == out'
            ]
      | out <- outputs
      ]

circuit :: Flash out -> FlashCircuit out
circuit Skip start = ([], finish)
  where
    finish = start

circuit (Emit out) start = ([(out, start)], finish)
  where
    finish = start

circuit Wait start = ([], finish)
  where
    finish = delay low start

circuit (IfThenElse cond (prog1, prog2)) start = (emits, finish)
  where
    (emits1, finish1) = circuit prog1 start1
    (emits2, finish2) = circuit prog2 start2

    start1 = and2 (start, cond)
    start2 = and2 (start, inv cond)

    emits  = emits1 ++ emits2
    finish = or2 (finish1, finish2)

circuit (While cond prog) start = (emits, finish)
  where
    (emits, finish') = circuit prog start'

    active   = or2 (start, finish')
    start'   = and2 (cond, active)
    finish   = and2 (inv cond, active)

circuit (prog1 :>> prog2) start = (emits, finish)
  where
    (emits1, finish1) = circuit prog1 start
    (emits2, finish)  = circuit prog2 finish1

    emits = emits1 ++ emits2

circuit (prog1 :|| prog2) start = (emits, finish)
  where
    (emits1, finish1) = circuit prog1 start
    (emits2, finish2) = circuit prog2 start

    emits  = emits1 ++ emits2
    finish = synchronize (finish1, finish2)        

----------------------------------------------------------------
-- parallel synchronization

synchronize :: (Signal Bool, Signal Bool) -> Signal Bool
synchronize (go1, go2) = go
  where
    both = and2 (go1, go2)
    one  = xor2 (go1, go2)
    wait = delay low (xor2 (one, wait))
    go   = or2 (both, and2 (wait, one))

-- this specification was used to verify the above implementation
-- correct

synchronizeSpec :: (Signal Bool, Signal Bool) -> (Signal Bool, Signal Bool)
synchronizeSpec (go1, go2) = (go, wrong)
  where
    [go,wrong] = compile program ["go", "wrong"]
    
    waiting =
      While (go1 <=> go2)
        ( IfThenElse (go1 <&> go2)
            ( Emit "go"
            , Skip
            )
      :>> Wait
        )
    
    program =
      While high
        ( waiting
      :>> IfThenElse go1
            ( Wait
          :>> waiting
          :>> IfThenElse go2
                ( Emit "go"
                , Emit "wrong"
                )
            , Wait
          :>> waiting
          :>> IfThenElse go1
                ( Emit "go"
                , Emit "wrong"
                )
            )
      :>> Wait
        )

prop_SynchronizeCorrect inp =
  always (inv wrong) ==>
    out1 <==> out2
 where
  (out1, wrong) = synchronizeSpec inp
  out2          = synchronize inp

always cond = ok
  where
    sofar = delay high ok
    ok    = and2 (cond, sofar)

----------------------------------------------------------------
-- the end.
