module Moo where

import Lava
import SequentialCircuits
import Vis
import List
  ( nub
  )

infix  5 :=
infixr 4 :>>
infixr 3 :||

----------------------------------------------------------------
-- Moo datatype

newtype Var
  = Var Int
 deriving (Eq, Show)

type Bit
  = Signal Bool

type Expr
  = (Var -> Bit) -> Bit

data Moo
  = New (Var -> Moo)
  | Var := Expr
  | Delay
  | Moo :>> Moo
  | Moo :|| Moo
  | IfThenElse Expr (Moo, Moo)
  | While Expr Moo
  | Repeat Moo Expr

data Program
  = Output (Var -> Program)
  | Start Moo

type Table
  = [(Var, (Bit, Bit))]

type State
  = (Bit, Table)

type Output
  = [Bit]

type Env
  = Table

type Collect
  = Table

type Circuit
  = (Env, State) -> (State, Collect)

----------------------------------------------------------------
-- operators

mergeT :: Table -> Table -> Table
mergeT table1 table2 =
  [ (var, (cha1 <|> cha2, ifThenElse cha1 (val1, val2)))
  | (var, (cha1, val1)) <- table1
  , let (cha2, val2) = lookUp var table2
  ]

mergeS :: State -> State -> State
mergeS state1 state2 =
  (active state1 <|> active state2, mergeT (updates state1) (updates state2))

updates :: State -> Table
updates (active, table) =
  [ (var, (active <&> change, val)) | (var, (change, val)) <- table ]

enable :: Bit -> State -> State
enable active' (active, table) = (active' <&> active, table)

synchronize :: State -> State -> State
synchronize = undefined

create :: Env -> Var
create env = Var (maximum (0 : map (ident . fst) env) + 1)
  where
    ident (Var i) = i

active :: State -> Bit
active (act,_) = act

empty :: State -> Collect
empty (active, tab) = [ (var,(low,low)) | (var,_) <- tab ]


removeT :: Var -> Table -> Table
removeT var table = filter ((var /=) . fst) table

removeS :: Var -> State -> State
removeS var (active, table) = (active, removeT var table)

lookUp :: Var -> Table -> (Bit,Bit)
lookUp var tab =
  case lookup var tab of
    Just val -> val
    Nothing  -> error ("no var '" ++ show var ++ "' in " ++ show tab)

value :: Expr -> State -> Bit
value expr (_, tab) = expr (\var -> snd (lookUp var tab))

update :: Var -> Bit -> State -> State
update var val (active, table) = (active,(var,(high,val)) : removeT var table)

----------------------------------------------------------------
-- compiling

compile :: Program -> Output
compile prog = output
  where
    (_, _, output) = global prog ([], (delay high low, []))
  
    global (Start moo) (env, state0) = (state2, coll, [])
      where
        (state2, coll) = circuit moo (env, state0)

    global (Output fprog) (env, state0) =
      variable (global . fprog) (env, state0)

variable :: (Var -> (Env, State) -> (State, Collect, Output))
         -> (Env, State) -> (State, Collect, Output)
variable f (env, state0) = (state2, coll, signal' : vs)
  where
    var                 = create env
    env1                = (var, (low,signal)) : env
    (state1, coll1, vs) = f var (env1, update var low state0)
    signal              = delay low signal'
    (req,val)           = lookUp var coll1
    signal'             = ifThenElse req (val,signal)
    state2              = removeS var state1
    coll                = removeT var coll1

circuit :: Moo -> Circuit
circuit (New fmoo) (env, state0) = (state1, coll)
  where
    (state1, coll, _)     = variable (dummy . circuit . fmoo) (env, state0)
    dummy f (env, state0) = (state, coll, [])
      where (state, coll) = f (env, state0)

circuit (var := expr) (env, state0) = (state1, empty state1)
  where
    state1          = update var (value expr state0) state0

circuit Delay (env, state0) = (state1, coll)
  where
    coll            = updates state0
    state1          = (delay low (active state0), env)

circuit (moo1 :>> moo2) (env, state0) = (state2, coll)
  where
    (state1, coll1) = circuit moo1 (env, state0)
    (state2, coll2) = circuit moo2 (env, state1)
    coll            = mergeT coll1 coll2

circuit (moo1 :|| moo2) (env, state0) = (state3, coll)
  where
    (state1, coll1) = circuit moo1 (env, state0)
    (state2, coll2) = circuit moo2 (env, state0)
    state3          = synchronize state1 state2
    coll            = mergeT coll1 coll2

circuit (IfThenElse expr (moo1, moo2)) (env, state0) = (state3, coll)
  where
    cond            = value expr state0
    (state1, coll1) = circuit moo1 (env, enable cond state0)
    (state2, coll2) = circuit moo2 (env, enable (inv cond) state0)
    state3          = mergeS state1 state2
    coll            = mergeT coll1 coll2

circuit (While expr moo) (env, state0) = (state3, coll)
  where
    state1          = mergeS state0 state2
    cond            = value expr state1
    (state2, coll)  = circuit moo (env, enable cond state1)
    state3          = enable (inv cond) state1

circuit (Repeat moo expr) (env, state0) = (state3, coll)
  where
    state1          = mergeS state0 (enable (inv cond) state2)
    cond            = value expr state2
    (state2, coll)  = circuit moo (env, state1)
    state3          = enable cond state2

----------------------------------------------------------------
-- the end.

